package net.shipilev.singleton;

import java.io.PrintWriter;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.TimeUnit;

public class Main {

    private static final int DURATION_SEC = Integer.getInteger("time", 10);
    private static final int THREADS = Integer.getInteger("threads", 2);

    public static void main(String[] args) throws BrokenBarrierException, InterruptedException {
        new Main().run();
    }

    private volatile Factory factory;
    private volatile boolean stopped;
    private final PrintWriter pw = new PrintWriter(System.err, true);

    public Singleton dummy;

    private void run() throws BrokenBarrierException, InterruptedException {
        System.err.println("Now using " + FactorySelector.newFactory() + " and " + SingletonSelector.newSingleton());

        // warmup to compile and inline everything
        for (int c = 0; c < 10000000; c++) {
            dummy = FactorySelector.newFactory().getInstance();
        }

        factory = FactorySelector.newFactory();

        long time1 = System.nanoTime();
        for (int c = 0; c < 10000000; c++) {
            dummy = factory.getInstance();
        }
        long time2 = System.nanoTime();
        pw.println("time: " + TimeUnit.NANOSECONDS.toMillis(time2 - time1));


        Runner[] runners = new Runner[THREADS];
        Thread[] threads = new Thread[THREADS];
        for (int t = 0; t < THREADS; t++) {
            runners[t] = new Runner();
            threads[t] = new Thread(runners[t]);
            threads[t].start();
        }

        new Thread(new Trigger()).start();

        for (int t = 0; t < THREADS; t++) {
            threads[t].join();
        }

        pw.flush();
    }

    public class Trigger implements Runnable {

        public void run() {
            try {
                TimeUnit.SECONDS.sleep(DURATION_SEC);
            } catch (InterruptedException e) {
                // fall-through
            }

            stopped = true;
        }
    }

    public class Runner implements Runnable {

        public void run() {
            long iterations = 0;
            long selfCheckFailed = 0;

            while (!stopped) {

                Singleton singleton = factory.getInstance();
                if (singleton == null || !singleton.selfCheck()) {
                    selfCheckFailed++;
                }
                iterations++;

                // switch to new factory
                factory = FactorySelector.newFactory();

            }

            pw.printf("%s: %d of %d had failed (%e)\n", Thread.currentThread().getName(), selfCheckFailed, iterations, 1.0D * selfCheckFailed / iterations);
        }
    }

}
